import unittest

# from bioimaging import epifm


class TestTirfm(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1(self):
        pass


if __name__ == '__main__':
    unittest.main()
